from redis import Redis

class SafeRedis(Redis):
    def __init__(self, *args, lock_timeout: int = 10, blocking_timeout: int = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self.lock_timeout = lock_timeout
        self.blocking_timeout = blocking_timeout

    def _get_lock(self, key):
        return self.lock(f"lock:{key}", timeout=self.lock_timeout)
    
    def safe_set(self, key, value, ex: int = None):
        lock = self._get_lock(key)
        if lock.acquire(blocking=True, blocking_timeout=self.blocking_timeout):
            try:
                self.set(key, value, ex=ex)
                return True
            finally:
                lock.release()
        else:
            return False
    
    def safe_delete(self, key):
        lock = self._get_lock(key)
        if lock.acquire(blocking=True, blocking_timeout=self.blocking_timeout):
            try:
                self.delete(key)
                return True
            finally:
                lock.release()
        else:
            return False


KEY_PREFIX = "_rqlite"

PRODUCER_LABEL = "producer"
WORKER_LABEL = "worker"
TOPIC_LABEL = "topic"
CHANNEL_LABEL = "channel"
SUBSCRIBER_LABEL = "subscriber"

META_SUFFIX = "meta"
DATA_SUFFIX = "data"

_get_topic_meta_key = lambda topic: f"{KEY_PREFIX}-{TOPIC_LABEL}-{topic}-{META_SUFFIX}"
_get_topic_partition_data_key = lambda topic, partition: f"{KEY_PREFIX}-{TOPIC_LABEL}-{topic}-{partition}-{DATA_SUFFIX}"
_get_topic_partition_worker_key = lambda topic, partition: f"{KEY_PREFIX}-{TOPIC_LABEL}-{topic}-{partition}-{WORKER_LABEL}"
_get_worker_topic_partition_key = lambda worker: f"{KEY_PREFIX}-{WORKER_LABEL}-{worker}-{TOPIC_LABEL}"
_get_lock_key = lambda key: f"{KEY_PREFIX}-lock:{key}"
_get_channel_subscriber_pre = lambda channel: f"{KEY_PREFIX}-{CHANNEL_LABEL}-{channel}-{SUBSCRIBER_LABEL}-"
_get_channel_subscriber_key = lambda channel, subscriber: f"{KEY_PREFIX}-{CHANNEL_LABEL}-{channel}-{SUBSCRIBER_LABEL}-{subscriber}"
_get_channel_data_subscriber_key = lambda channel, subscriber: f"{KEY_PREFIX}-{CHANNEL_LABEL}-{channel}-data-{subscriber}"


from redis import Redis
import json

from .config import _get_topic_partition_data_key, _get_channel_subscriber_pre
from .utils import create_topic, get_topic_meta, _get_partition
from .redis import SafeRedis

class RQProducer:
    """
    RQLite Producer:
    """
    def __init__(self, 
                 host: str = "localhost", 
                 port: int = 6379, 
                 username: str = None, 
                 password: str = None,
                 redis_conn: SafeRedis = None,
                 serializer = None):
        
        """
        Initialize the RQLite Producer with 
        - host: Hostname for Redis server
        - port: Port where redis server is listening
        - username: Username of Redis server (if any)
        - password: Password of Redis server (if any)
        - redis_conn: Redis Connection object from Redis library
        - serializer: Serializer class with serialize method for the message
        """
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.serializer = serializer

        if redis_conn:
            self.conn = redis_conn
        else:
            self.conn = SafeRedis(host=host, port=port, username=username, password=password)

        if not self._test_conn():
            raise Exception(f"Could not connect to Redis at {host}:{port}.")
    
    def _test_conn(self):
        """
        Implements test redis connection.
        """
        return self.conn.ping()
    
    def send(self, topic: str, message, partition_key = None):
        """
        Implements send message to Redis list with the name _rqlite-topic-{topic}-{partition}-data.
        Topic is created if topic does not exist.
        Partition is decided based on partition_key. Default partition is 0.
        Message is RPUSH'ed to Redis list
        """
        try:
            if self.serializer:
                message = self.serializer.serialize(message)
        except Exception as ex:
            raise Exception(f"Error serializing the message '{message}'")
        
        topic_metadata: dict = get_topic_meta(self.conn, topic)
        if not topic_metadata:
            topic_metadata = create_topic(self.conn, topic)
        
        if not topic_metadata:
            raise Exception(f"Error creating topic '{topic}' in Redis.")
        
        partition = _get_partition(partition_key, topic_metadata.get("partitions", 1))

        self.conn.rpush(_get_topic_partition_data_key(topic, partition), message)     
        
        return True


    def broadcast(self, channel: str, message):
        """
        Implements broadcast message to all Redis lists corresponding to subscriber with the prefix _rqlite-channel-{channel}-subscriber.
        Message is RPUSH'ed to Redis list
        """

        try:
            if self.serializer:
                message = self.serializer.serialize(message)

        except Exception as ex:
            raise Exception(f"Error serializing the message '{message}")

        channel_subscriber_keys = [key.decode("utf-8") for key in self.conn.keys(f"{_get_channel_subscriber_pre(channel)}*")]

        for subscriber_key in channel_subscriber_keys:
            channel_data_subscriber_key = self.conn.get(subscriber_key)
            if not channel_data_subscriber_key:
                continue
            channel_data_subscriber_key = channel_data_subscriber_key.decode("utf-8")
            self.conn.rpush(channel_data_subscriber_key, message)

        return True




from redis import Redis
import json
import hashlib
from uuid import uuid4
from typing import Callable

from .config import _get_topic_meta_key


def create_topic(redis_conn: Redis, topic: str, num_partitions: int = 1):
    metadata = {
        "name": topic,
        "partitions": num_partitions
    }
    try:
        redis_conn.set(_get_topic_meta_key(topic), json.dumps(metadata))
    except:
        raise Exception(f"ERROR: Could not create RQLite topic '{topic}' in Redis")
    
    return metadata

def get_topic_meta(redis_conn: Redis, topic: str):
    topic_metadata: bytes = redis_conn.get(_get_topic_meta_key(topic))
    if topic_metadata:
        topic_metadata = json.loads(topic_metadata.decode())
    return topic_metadata

def _hash_md5(key: str):
    if not key:
        raise ValueError(f"ERROR: Invalid key '{key}'")
    
    return int(hashlib.md5(key.encode("utf-8")).hexdigest(), 16)

def _get_partition(partition_key: str, num_partitions: int, hash_func: Callable[[str], int] = _hash_md5):
    if not partition_key:
        return 0
    partition_key = str(partition_key)
    return hash_func(partition_key) % num_partitions
    
def _gen_id():
    return hashlib.md5(str(uuid4()).encode("utf-8")).hexdigest()


from redis import Redis
from typing import Callable
import threading
import time
import logging
import traceback

from .utils import _gen_id
from .config import _get_channel_subscriber_key, _get_channel_data_subscriber_key


class RQSubscriber:
	def __init__(self,
				host: str = "localhost",
				port: int = 6379,
				username: str = None,
				password: str = None,
				redis_conn: Redis = None,
				channel: str = None,
				callback: Callable = None,
				serializer = None,
				timeout_ms: int = 10000,
				poll_timeout_ms: int = 1000):
		"""
		Initialize RQLite Subscriber
		- host: Hostname for Redis server
		- port: Port where Redis server is listening
		- username: Username of Redis server (if any)
		- password: Password of Redis server (if any)
		- redis_conn: Redis Connection object from Redis library
		- channel: Name of the channel to subscribe
		- callback: Callback function executed on message poll
		- serializer: Serializer class with deserialize method for the message
        - timeout_ms: Timeout for the worker (in milliseconds)
        - poll_timeout_ms: Sleep time (in milliseconds) for polling if no messages available
		"""

		if not channel:
			raise Exception(f"Error: RQlite subscriber channel cannot be None")

		if poll_timeout_ms >= timeout_ms:
			raise Exception(f"ERROR: RQlite Subscriber's poll_timeout_ms must be less than timeout_ms")

		self.id = _gen_id()
		self.channel = channel
		self.callback = callback
		self.serializer = serializer

		if redis_conn:
			self.conn = redis_conn
		else:
			self.conn = Redis(host=host, port=port, username=username, password=password)

		self.timeout_ms = timeout_ms
		self.poll_timeout_ms = poll_timeout_ms

		self._thread = None
		self._running = None
		self._processing = None


	def poll(self):
		"""
		Impelments polling message from subscriber specific data queues
		"""
		subscriber_key = _get_channel_subscriber_key(self.channel, self.id)
		data_key = self.conn.get(subscriber_key)
		if not data_key:
			data_key = self._spawn()
		else:
			data_key = data_key.decode("utf-8")

		msg = self.conn.lpop(data_key)

		if not msg:
			return None

		if self.serializer:
			msg = self.serializer.deserialize(msg)

		return msg


	def _spawn(self):
		"""
		Implements setting RQLite Subscriber key to Redis Server for broadcast to find
		"""
		subscriber_key = _get_channel_subscriber_key(self.channel, self.id)
		data_key = _get_channel_data_subscriber_key(self.channel, self.id)

		self.conn.set(subscriber_key, data_key, ex=int(self.timeout_ms/1000))
		return data_key

	def _remove(self):
		"""
		Impelments removing RQLite Subscriber key from Redis Server
		"""
		subscriber_key = _get_channel_subscriber_key(self.channel, self.id)
		self.conn.delete(subscriber_key)
		return True


	def beat(self):
		"""
		Implements sending heartbeat to keep RQLite Subscriber alive 
		"""
		subscriber_key = _get_channel_subscriber_key(self.channel, self.id)
		is_key = self.conn.expire(subscriber_key, int(self.timeout_ms/1000))
		if not is_key:
			raise Exception(f"ERROR: Cannot send heartbeat for RQLite subscriber '{self.id}'. Subscriber has timed out")

		return True


	def commit(self):
		"""
		Implements commit for consuming a message.
		Currently only implements sending a heartbeat
		"""
		self.beat()


	def run(self):
		while self._running:
		    try:
		        msg = self.poll()

		        if not msg:
		            time.sleep(self.poll_timeout_ms / 1000.)
		            continue

		        self._processing = True
		        if self.callback:
		            self.callback(msg, heartbeat=self.beat)
		        logging.info(f"Subscriber '{self.id}'' finished processing message from channel '{self.channel}'")
		    except Exception as ex:
		        traceback.print_exc()
		        logging.error(f"{type(ex).__name__}: {str(ex)}. Failed to run callback function '{self.callback.__name__}' for message '{msg}'.")
		    finally:
		        self.commit()
		        self._processing = False



	def start(self):
		"""
		Start subscriber on a background thread 
		"""
		logging.info(f"Staring RQLite Subscriber '{self.id}' listening on channel '{self.channel}'")
		self._thread = threading.Thread(target=self.run, daemon=True)
		self._running = True
		self._thread.start()


	def stop(self, safe: bool = True):
		"""
        Stop subscriber thread
        - safe: Waits for current message to finish processing and remove subscriber before stopping.
        """
		logging.info(f"Stopping RQLite Subscriber '{self.id}' listening on channel '{self.channel}'")
		self._running = False
		if safe:
		    self._thread.join()
		self._remove()



from redis import Redis
from typing import Callable
import threading
import time
import logging
import traceback

from .utils import _gen_id, get_topic_meta, create_topic
from .config import _get_topic_partition_data_key, _get_topic_partition_worker_key, _get_worker_topic_partition_key, _get_lock_key

class RQWorker:
    def __init__(self, 
                host: str = "localhost",
                port: int = 6379, 
                username: str = None, 
                password: str = None, 
                redis_conn: Redis = None, 
                topic: str = None, 
                callback: Callable = None, 
                serializer = None,
                timeout_ms: int = 10000,
                poll_timeout_ms: int = 1000):
        """
        Initialize RQLite Worker
        - host: Hostname for Redis server
        - port: Port where redis server is listening
        - username: Username of Redis server (if any)
        - password: Password of Redis server (if any)
        - redis_conn: Redis Connection object from Redis library
        - topic: Name of the topic to listen
        - callback: Callback function to execute on the message polled
        - serializer: Serializer class with deserialize method for the message
        - timeout_ms: Timeout for the worker in milliseconds
        - poll_timeout_ms: Sleep time for polling if no messages available
        """
        if not topic:
            raise Exception(f"Error: RQLite worker topic cannot be None")
        
        self.id = _gen_id()
        self.topic = topic        
        self.callback = callback
        self.serializer = serializer
        
        if redis_conn:
            self.conn = redis_conn
        else:
            self.conn = Redis(host=host, port=port, username=username, password=password)

        self.timeout_ms = timeout_ms
        self.poll_timeout_ms = poll_timeout_ms

        self._partition = None
        self._thread = None
        self._running = False
        self._processing = False

    
    def poll(self):
        """
        Implements polling message from the topic queues
        """
        topic_metadata = get_topic_meta(self.conn, self.topic)
        if not topic_metadata:
            topic_metadata = create_topic(self.conn, self.topic)
        
        # Fetch number of messages in all partition queues
        if self._partition is not None:
            self.beat()
            message = self.conn.lpop(_get_topic_partition_data_key(self.topic, self._partition))
            if not message:
                self._detach()
                return self.poll()
            
            if self.serializer:
                message = self.serializer.deserialize(message)
            
            return message
            
        else:
            for i in range(topic_metadata.get("partitions")):
                if self.conn.get(_get_topic_partition_worker_key(self.topic, i)):
                    continue
                partition_size = self.conn.llen(_get_topic_partition_data_key(self.topic, i))
                if partition_size == 0:
                    continue

                if not self._attach(i):
                    continue

                message = self.conn.lpop(_get_topic_partition_data_key(self.topic, self._partition))
                if not message:
                    self._detach()
                    continue

                if self.serializer:
                    message = self.serializer.deserialize(message)
                
                return message

        return None

    def _detach(self):
        """
        Detach worker from the current partition. 
        """

        # self.conn.delete(_get_worker_topic_partition_key(self.id))
        topic_partition_key = _get_topic_partition_worker_key(self.topic, self._partition)
        lock = self.conn.lock(_get_lock_key(topic_partition_key), timeout=10)
        if lock.acquire(blocking=True, blocking_timeout=1):
            worker_id_bytes: bytes = self.conn.get(topic_partition_key)
            if worker_id_bytes and worker_id_bytes.decode("utf-8") == self.id:
                self.conn.delete(topic_partition_key)
            lock.release()
        
        self._partition = None

    def _attach(self, partition: int): 
        """
        Attach worker with the given partition. 
        """
        is_attached = False
        topic_partition_key = _get_topic_partition_worker_key(self.topic, partition)
        lock = self.conn.lock(_get_lock_key(topic_partition_key), timeout=10)
        if lock.acquire(blocking=True, blocking_timeout=1):
            worker_id_bytes: bytes = self.conn.get(topic_partition_key)
            if not worker_id_bytes:
                self.conn.set(topic_partition_key, self.id, ex=int(self.timeout_ms/1000))
                is_attached = True
            lock.release()

        if not is_attached:
            return False

        # self.conn.set(_get_worker_topic_partition_key(self.id), f"{self.topic}-{partition}", ex=int(self.timeout_ms * 1000))
        self._partition = partition
        return True

    def beat(self):
        """
        Implements sending heartbeat to keep the RQLite worker alive
        """
        topic_partition_key = _get_topic_partition_worker_key(self.topic, self._partition)
        lock = self.conn.lock(_get_lock_key(topic_partition_key), timeout=10)
        if lock.acquire(blocking=True, blocking_timeout=1):
            worker_id_bytes: bytes = self.conn.get(topic_partition_key)
            if worker_id_bytes:
                if worker_id_bytes.decode("utf-8") == self.id:
                    self.conn.set(topic_partition_key, self.id, ex=int(self.timeout_ms/1000))
                else:
                    self._partition = None
                    raise Exception(f"ERROR: Cannot send beat for RQLite worker '{self.id}' on queue '{self.topic}-{self._partition}'. Queue attached to a different worker.")
            else:
                self._partition = None
                raise Exception(f"ERROR: Cannot send heartbeat for RQLite worker '{self.id}' on queue '{self.topic}-{self._partition}'. Worker has timed out.")
            lock.release()
        else:
            raise Exception(f"ERROR: Cannot obtain lock on key '{topic_partition_key}' in Redis.")

        return True

    def commit(self):
        """
        Implements commit for consuming the message. 
        Currently only implements sending a heartbeat
        """
        if not self._partition:
            return
        self.beat()
        
    def run(self):
        while self._running:
            try:
                msg = self.poll()

                if not msg:
                    time.sleep(self.poll_timeout_ms / 1000.)
                    continue

                self._processing = True
                if self.callback:
                    self.callback(msg, heartbeat=self.beat)
                logging.info(f"Worker {self.id} finished processing message from queue {self.topic}-{self._partition}")
            except Exception as ex:
                traceback.print_exc()
                logging.error(f"{type(ex).__name__}: {str(ex)}. Failed to run callback function '{self.callback.__name__}' for message '{msg}'.")
            finally:
                self.commit()
                self._processing = False

    def start(self):
        """
        Start worker on a background thread
        """
        logging.info(f"Staring RQLite Worker '{self.id}' listening on topic '{self.topic}'")
        self._thread = threading.Thread(target=self.run, daemon=True)
        self._running = True
        self._thread.start()

    def stop(self, safe: bool = True):
        """
        Stop worker thread
        - safe: Waits for current message to finish processing and detach worker before stopping.
        """
        logging.info(f"Stopping RQLite Worker '{self.id}' listening on topic '{self.topic}'")
        self._running = False
        if safe:
            self._thread.join()
        self._detach()


class RQWorkerPool:
    def __init__(self, 
                host: str = "localhost",
                port: int = 6379, 
                username: str = None, 
                password: str = None, 
                redis_conn: Redis = None, 
                topic: str = None, 
                callback: Callable = None, 
                serializer = None,
                timeout_ms: int = 10000,
                poll_timeout_ms: int = 1000, 
                num_workers: int = 1):
        """
        Initialize RQLite Worker Pool
        - host: Hostname for Redis server
        - port: Port where redis server is listening
        - username: Username of Redis server (if any)
        - password: Password of Redis server (if any)
        - redis_conn: Redis Connection object from Redis library
        - topic: Name of the topic to listen
        - callback: Callback function to execute on the message polled
        - serializer: Serializer class with deserialize method for the message
        - timeout_ms: Timeout for the worker in milliseconds
        - poll_timeout_ms: Sleep time for polling if no messages available
        - num_workers: Number of concurrent workers on the same topic
        """
        if not topic:
            raise Exception(f"Error: RQLite worker topic cannot be None")
        
        self.id = _gen_id()
        self.topic = topic        
        self.callback = callback
        self.serializer = serializer
        
        if redis_conn:
            self.conn = redis_conn
        else:
            self.conn = Redis(host=host, port=port, username=username, password=password)

        self.timeout_ms = timeout_ms
        self.poll_timeout_ms = poll_timeout_ms

        self._running = False

        self.workers: list[RQWorker] = [RQWorker(redis_conn=self.conn, topic=topic, callback=callback, serializer=serializer, timeout_ms=timeout_ms, poll_timeout_ms=poll_timeout_ms) for _ in range(num_workers)]

    def start(self):
        logging.info(f"Starting RQLite worker pool '{self.id}' listening on topic '{self.topic}'")
        self._running = True
        for worker in self.workers:
            worker.start()
    
    def stop(self):
        logging.info(f"Stopping RQLite worker pool '{self.id}' listening on topic '{self.topic}'")
        for worker in self.workers:
            worker.stop()
        self._running = False